
#include "files.h"

double swap(const double &x) {
  typedef union {
    unsigned char b[8];
    double d;
  } u64;
  u64 u;

  u.b[0] = ((u64*)&x)->b[7];
  u.b[1] = ((u64*)&x)->b[6];
  u.b[2] = ((u64*)&x)->b[5];
  u.b[3] = ((u64*)&x)->b[4];
  u.b[4] = ((u64*)&x)->b[3];
  u.b[5] = ((u64*)&x)->b[2];
  u.b[6] = ((u64*)&x)->b[1];
  u.b[7] = ((u64*)&x)->b[0];
	
  return u.d;
}

long swap(const long &ref) {
  return ((((unsigned char*)&ref)[0] << 24) | 
          (((unsigned char*)&ref)[1] << 16) | 
          (((unsigned char*)&ref)[2] <<  8) | 
          (((unsigned char*)&ref)[3] <<  0));
}

unsigned long swap(const unsigned long &ref) {
  return ((((unsigned char*)&ref)[0] << 24) | 
          (((unsigned char*)&ref)[1] << 16) | 
          (((unsigned char*)&ref)[2] <<  8) | 
          (((unsigned char*)&ref)[3] <<  0));
}

int swap(const int &ref) {
  if (sizeof(int)==4) return swap((long)ref);
  else return swap((short)ref);
}

unsigned int swap(const unsigned int &ref) {
  if (sizeof(int)==4) return swap((unsigned long)ref);
  else return swap((unsigned short)ref);
}

unsigned short swap(const unsigned short &ref) {
  return ((((unsigned char*)&ref)[0] << 8) | 
          (((unsigned char*)&ref)[1] << 0));
}

short swap(const short &ref) {
  return ((((unsigned char*)&ref)[0] << 8) | 
          (((unsigned char*)&ref)[1] << 0));
}

unsigned char swap(const unsigned char &ref) {
  return ref;
}

char swap(const char &ref) {
  return ref;
}

CDiskFile::CDiskFile() {
  fp = NULL;
}

CDiskFile::~CDiskFile() {
  Close();
}

void CDiskFile::OpenRead(const char *name) {
  Close();
  fp = fopen(name,"rb");
  if (!fp) THROW(FILE_OPEN_ERROR);
  GetSize();
}

void CDiskFile::OpenRewrite(const char *name) {
  Close();
  fp = fopen(name,"wb");
  if (!fp) THROW(FILE_OPEN_ERROR);
}

void CDiskFile::Close() {
  if (fp) fclose(fp); fp = NULL;
}

void CDiskFile::Read(void *buffer, unsigned long length) {
  if (length!=fread(buffer,1,length,fp)) THROW(FILE_READ_ERROR);
}

bool CDiskFile::ReadLn(char *str, int max) {
  if (!fgets(str,max,fp)) {
    if (feof(fp)) return false;
    THROW(FILE_READ_ERROR);
  }
  int len = strlen(str);
  if ((len>0) && (str[len-1]==0x0A)) len--;
  if ((len>0) && (str[len-1]==0x0D)) len--;
  str[len]='\0';
  return true;
}


void CDiskFile::Write(const void *buffer, unsigned long length) {
  if (length!=fwrite(buffer,1,length,fp)) THROW(FILE_WRITE_ERROR);
}

void CDiskFile::PrintLn(const char* str) {
  Write(str,strlen(str));
  Write("\x0D\x0A",2);
}

void CDiskFile::GetSize() {
  if (0!=fseek(fp, 0, SEEK_END)) THROW(FILE_SEEK_ERROR);
  filesize = ftell(fp);
  Seek(0);
}

void CDiskFile::Seek(long pos) {
  if (0!=fseek(fp, pos, SEEK_SET)) THROW(FILE_SEEK_ERROR);
}

long CDiskFile::Size() const {
  return filesize;
}

const char* CDiskFile::ExceptionString(exception err) {
  switch (err) {
    case FILE_OPEN_ERROR:    return "File open error";
    case FILE_READ_ERROR:    return "File read error";
    case FILE_WRITE_ERROR:   return "File write error";
    case FILE_SEEK_ERROR:    return "File seek error";
    case FILE_FORMAT_ERROR:  return "Wrong file format";
    case FILE_COMPILE_ERROR: return "Struct size mismatch";
    default: return NULL;
  }
};

#ifndef FILES_NO_PDBFILE

CPDBfile::CPDBfile() {
  resources = NULL; records = NULL;
}

CPDBfile::~CPDBfile() {
  Close();
}

void CPDBfile::OpenRead(const char *name) {
  if (sizeof(PDBhead)!=SIZEOF_PDBHEAD) THROW(FILE_COMPILE_ERROR);
  if (sizeof(PDBResourceIndex)!=SIZEOF_PDBRESOURCEINDEX) THROW(FILE_COMPILE_ERROR);
  if (sizeof(PDBRecordIndex)!=SIZEOF_PDBRECORDINDEX) THROW(FILE_COMPILE_ERROR);

  CDiskFile::OpenRead(name);
  Read(&head,SIZEOF_PDBHEAD);
  entries = BIG(head.entries);
  if (ResDB()) { 
    resources = new PDBResourceIndex[entries];
    Read(resources,entries*SIZEOF_PDBRESOURCEINDEX);
  }
  else {
    records = new PDBRecordIndex[entries];
    Read(records,entries*SIZEOF_PDBRECORDINDEX);
  }
}

void CPDBfile::OpenRewrite(const char *name, const char *dbname, long type, long creator, unsigned short entries, bool res) {
  CDiskFile::OpenRewrite(name);
  memset(&head,0,sizeof(head));
  strcpy(head.name,dbname);
  time_t unixtime; time(&unixtime);
  head.flags = BIG((short)(res ? 1 : 0));
  head.version = BIG((short)1);
  head.crdate = head.moddate = head.backupdate = BIG((long)(unixtime+2082844800L));
  head.modnum = BIG((long)0);
  head.app_info_offset = BIG((long)0);
  head.sort_info_offset = BIG((long)0);
  head.type = BIG(type);
  head.creator = BIG(creator);
  head.unique_id_seed = BIG((long)0);
  head.next_record_list_id = BIG((long)0);
  head.entries = BIG(entries);
  Write(&head,SIZEOF_PDBHEAD);
  // PREPARE INDEXES
  WriteIndexPrepare = WriteIndexFinal = 0;
  WriteOffset = SIZEOF_PDBHEAD + 2;
  if (res) {
    WriteOffset += entries*SIZEOF_PDBRESOURCEINDEX;
    resources = new PDBResourceIndex[entries];
  }
  else {
    WriteOffset += entries*SIZEOF_PDBRECORDINDEX;
    records = new PDBRecordIndex[entries];
  }
  filesize = WriteOffset;
  this->entries = BIG(head.entries);
}

void CPDBfile::WriteRecordInfo(unsigned short size) {
  if (!records) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexPrepare>=entries) THROW(FILE_FORMAT_ERROR);
  PDBRecordIndex info;
  info.offset = BIG(WriteOffset);
  info.attr   = 0;
  memset(&info.uid,0,3);
  records[WriteIndexPrepare++] = info;
  WriteOffset += size;
  filesize += size;
  if (WriteIndexPrepare>=entries) {
    Write(records,entries*SIZEOF_PDBRECORDINDEX);
    short zero = 0;
    Write(&zero,2);
  }
}

void CPDBfile::WriteResourceInfo(unsigned short size,long type, unsigned short id) {
  if (!resources) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexPrepare>=entries) THROW(FILE_FORMAT_ERROR);
  PDBResourceIndex info;
  info.offset = BIG(WriteOffset);
  info.type   = BIG(type);
  info.id     = BIG(id);
  resources[WriteIndexPrepare++] = info;
  WriteOffset+= size;
  filesize += size;
  if (WriteIndexPrepare>=entries) {
    Write(resources,entries*sizeof(PDBResourceIndex));
    short zero = 0;
    Write(&zero,2);
  }
}

void CPDBfile::WriteRecord(void *buffer) {
  if (!records) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexPrepare<entries) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexFinal>=entries) THROW(FILE_FORMAT_ERROR);
  Write(buffer,GetRecordSize(WriteIndexFinal++));
}

void CPDBfile::WriteResource(void *buffer) {
  if (!records) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexPrepare<entries) THROW(FILE_FORMAT_ERROR);
  if (WriteIndexFinal>=entries) THROW(FILE_FORMAT_ERROR);
  Write(buffer,GetResourceSize(WriteIndexFinal++));
}

void CPDBfile::Close() {
  CDiskFile::Close();
  if (resources) delete resources; resources = NULL;
  if (records) delete records; records = NULL;
}

bool CPDBfile::ResDB() const {
  return ((BIG(head.flags) & 0x01)!=0);
}

unsigned short CPDBfile::Entries() const {
  return entries;
}

const char* CPDBfile::Name() const {
  return head.name;
}

long CPDBfile::Creator() const {
  return BIG(head.creator);
}

long CPDBfile::Type() const {
  return BIG(head.type);
}

void CPDBfile::GetResourceInfo(unsigned short index, ResourceInfo &info) const {
  if (!resources || (index>=entries)) THROW(FILE_FORMAT_ERROR);
  PDBResourceIndex &ref = resources[index];
  info.type   = BIG(ref.type);
  info.id     = BIG(ref.id);
  info.offset = BIG(ref.offset);
  if (index==entries-1) 
    info.size = (unsigned short)(Size()-info.offset);
  else
    info.size = (unsigned short)(BIG(resources[index+1].offset)-info.offset);
}

void CPDBfile::GetRecordInfo(unsigned short index, RecordInfo &info) const {
  if (!records || (index>=entries)) THROW(FILE_FORMAT_ERROR);
  PDBRecordIndex &ref = records[index];
  info.offset = BIG(ref.offset);
  info.attr   = ref.attr;
  if (index==entries-1) 
    info.size = (unsigned short)(Size()-info.offset);
  else
    info.size = (unsigned short)(BIG(records[index+1].offset)-info.offset);
}

bool CPDBfile::FindResource(long type, unsigned short id, unsigned short &index) const {
  ResourceInfo res;
  for (int i=0; i<Entries(); i++) {
    GetResourceInfo(i,res);
    if ((res.type==type) && (res.id==id)) {
      index = i;
      return true;
    }
  }
  return false;
}

unsigned short CPDBfile::GetResourceSize(unsigned short index) const {
  ResourceInfo info;
  GetResourceInfo(index,info);
  return info.size;
}

void CPDBfile::ReadResource(unsigned short index, void *buffer) {
  ResourceInfo info;
  GetResourceInfo(index,info);
  Seek(info.offset);
  Read(buffer,info.size);
}

unsigned short CPDBfile::GetRecordSize(unsigned short index) const {
  RecordInfo info;
  GetRecordInfo(index,info);
  return info.size;
}

void CPDBfile::ReadRecord(unsigned short index, void *buffer) {
  RecordInfo info;
  GetRecordInfo(index,info);
  Seek(info.offset);
  Read(buffer,info.size);
}

#endif

